  /*                      \\\|///
     \\  - -  //
     (  @ @  )
     /--------------------oOOo-(_)-oOOo---------------------\
     |                                                      |
     |                                                      |
     |    Transistor Tester for Arduino (version 1.08a)     |
     |                                                      |
     |  based on code: Karl-Heinz Kubbeler (version 1.08k)  |
     |                                                      |
     |                                                      |
     |                            Oooo                      |
     \--------------------oooO----(   )---------------------/
     (   )    ) /
     \ (    (_/
     \_)                            */
// tradución seta43 para ARDUINO Nano Español

#include <avr/io.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <string.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <math.h>
#include <stdint.h>
#include <avr/power.h>




const uint16_t RLtab[] PROGMEM =
    { 22447, 20665, 19138, 17815, 16657, 15635, 14727, 13914, 13182, 12520, 11918, 11369, 10865, 10401, 9973, 9577, 9209, 8866, 8546, 8247, 7966,
7702, 7454, 7220, 6999, 6789, 6591, 6403, 6224, 6054, 5892, 5738, 5590, 5449, 5314, 5185, 5061, 4942, 4828, 4718, 4613, 4511, 4413, 4319, 4228 };
const uint16_t RHtab[] PROGMEM = { 954, 903, 856, 814, 775, 740, 707, 676, 648 };

const unsigned char TestRunning[] PROGMEM = "---------------";
const unsigned char BatWeak[] PROGMEM = "debil";
const unsigned char BatEmpty[] PROGMEM = "vacio!";
const unsigned char TestFailed2[] PROGMEM = "estropeado";
const unsigned char Component[] PROGMEM = " parte";
const unsigned char Triac[] PROGMEM = "Triac";
const unsigned char Thyristor[] PROGMEM = "Tiristor";
const unsigned char Unknown[] PROGMEM = " esconocido";
const unsigned char TestFailed1[] PROGMEM = "No, desconocido, o";
const unsigned char OrBroken[] PROGMEM = "o estropeado ";
const unsigned char TestTimedOut[] PROGMEM = "Fuera de tiempo!";
const unsigned char Bat_str[] PROGMEM = "Bat. ";
const unsigned char OK_str[] PROGMEM = "OK";
const unsigned char mosfet_str[] PROGMEM = "-MOS";
const unsigned char jfet_str[] PROGMEM = "JFET";
const unsigned char GateCap_str[] PROGMEM = "C=";
const unsigned char hfe_str[] PROGMEM = "B=";
const unsigned char NPN_str[] PROGMEM = "NPN ";
const unsigned char PNP_str[] PROGMEM = "PNP ";
const unsigned char Ir_str[] PROGMEM = "  Ir=";

const unsigned char N123_str[] PROGMEM = " 123=";
const unsigned char Uf_str[] PROGMEM = "Uf=";
const unsigned char vt_str[] PROGMEM = " Vt=";
const unsigned char Vgs_str[] PROGMEM = "@Vgs=";
const unsigned char CapZeich[] PROGMEM = { '-', 3, 3, '-', 0 };

const unsigned char Cell_str[] PROGMEM = "Cell!";
const unsigned char VCC_str[] PROGMEM = "VCC=";
const unsigned char VERSION_str[] PROGMEM = "Test_COMPONENTE";
const unsigned char AnKat[] PROGMEM = { '-', 1, 3, '-', 0 };
const unsigned char KatAn[] PROGMEM = { '-', 3, 2, '-', 0 };
const unsigned char Diodes[] PROGMEM = { '*', 1, ' ', ' ', 0 };
const unsigned char Resistor_str[] PROGMEM = { '-', 7, 0x3D, 6, '-', 0 };
const unsigned char DiodeIcon1[] PROGMEM = { 0x11, 0x19, 0x1d, 0x1f, 0x1d, 0x19, 0x11, 0x00 };
const unsigned char DiodeIcon2[] PROGMEM = { 0x11, 0x13, 0x17, 0x1f, 0x17, 0x13, 0x11, 0x00 };
const unsigned char CapIcon[] PROGMEM = { 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x00 };
const unsigned char ResIcon1[] PROGMEM = { 0x00, 0x0f, 0x08, 0x18, 0x08, 0x0f, 0x00, 0x00 };
const unsigned char ResIcon2[] PROGMEM = { 0x00, 0x1e, 0x02, 0x03, 0x02, 0x1e, 0x00, 0x00 };
const unsigned char OmegaIcon[] PROGMEM = { 0x00, 0x00, 0x0e, 0x11, 0x11, 0x0a, 0x1b, 0x00 };
const unsigned char MicroIcon[] PROGMEM = { 0x00, 0x00, 0x0a, 0x0a, 0x0a, 0x0e, 0x09, 0x10 };
const unsigned char PinRLtab[] PROGMEM = { (1 << (0 * 2)), (1 << (1 * 2)), (1 << (2 * 2)) };
const unsigned char PinADCtab[] PROGMEM = { (1 << 0), (1 << 1), (1 << 2) };
const uint8_t PrefixTab[] PROGMEM = { 'p', 'n', 228, 'm', 0, 'k', 'M' };
const uint8_t EE_ESR_ZEROtab[] PROGMEM = { 20, 20, 20, 20 };

unsigned int RHmultip = 884;





struct Diode_t {
    uint8_t Anode;
    uint8_t Cathode;
    unsigned int Voltage;
};
struct Diode_t diodes[6];
uint8_t NumOfDiodes;
struct {
    unsigned long hfe[2];
    unsigned int uBE[2];
    uint8_t b, c, e;
} trans;
unsigned int gthvoltage;
uint8_t PartReady;
uint8_t PartMode;
uint8_t tmpval, tmpval2;
unsigned int ref_mv;
struct resis_t {
    unsigned long rx;
    uint8_t ra, rb;
    uint8_t rt;
} resis[3];
uint8_t ResistorsFound;
uint8_t ii;
struct cap_t {
    unsigned long cval;
    unsigned long cval_max;
    union t_combi {
	unsigned long dw;
	uint16_t w[2];
    } cval_uncorrected;
    uint8_t ca, cb;
    int8_t cpre;
    int8_t cpre_max;
} cap;
volatile uint16_t ovcnt16;
volatile uint8_t unfinished;
int16_t load_diff;
uint8_t WithReference;
uint8_t PartFound;
char outval[12];
uint8_t empty_count;
uint8_t mess_count;
struct ADCconfig_t {
    uint8_t Samples;
    uint8_t RefFlag;
    uint16_t U_Bandgap;
    uint16_t U_AVCC;
} ADCconfig;
EMPTY_INTERRUPT(TIMER2_COMPA_vect);
EMPTY_INTERRUPT(ADC_vect);
uint8_t tmp = 0;
byte TestKey;
byte TestKeyPin = 17;




void setup()
{

    Serial.begin(9600);

    pinMode(TestKeyPin, INPUT);

    lcd_pgm_custom_char(1, DiodeIcon1);
    lcd_pgm_custom_char(2, DiodeIcon2);
    lcd_pgm_custom_char(3, CapIcon);
    lcd_pgm_custom_char(7, ResIcon1);
    lcd_pgm_custom_char(6, ResIcon2);
    lcd_pgm_custom_char(244, OmegaIcon);
    lcd_pgm_custom_char(228, MicroIcon);
/*
    lcd_string("TransistorTesterGambas");
    lcd_set_cursor(1, 0);
    lcd_string("ArduinoGambas 1.08a");
*/
    ADCSRA = (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
    if (tmp) 
    {
	lcd_set_cursor(0, 0);
	lcd_fix_string(TestTimedOut);
	sleep_5ms(600);
    }
    MCUCR = (1 << PUD);
    empty_count = 0;
    mess_count = 0;
}



void loop()
{
  start:
    TestKey = 1;
    while (TestKey) 
    {
	TestKey = digitalRead(TestKeyPin);
	delay(100);
    }
    while (!TestKey) 
    {
	TestKey = digitalRead(TestKeyPin);
	delay(100);
    }


    lcd_clear();
    delay(100);
    PartFound = 0;
    NumOfDiodes = 0;
    PartReady = 0;
    PartMode = 0;
    WithReference = 0;
    DDRC = 0xF8;
    ResistorsFound = 0;
    cap.ca = 0;
    cap.cb = 0;
    ADCconfig.RefFlag = 0;
    Calibrate_UR();
    lcd_set_cursor(0, 0);
    ADCconfig.U_Bandgap = (ref_mv + 3);
    lcd_pgm_string(VERSION_str);
    RefVoltage();
    lcd_set_cursor(1, 0);
    lcd_fix_string(TestRunning);
    lcd_set_cursor(1, 0);
    delay(5);
    EntladePins();
    if (PartFound == 8) 
		{
		lcd_clear();
		lcd_fix_string(Cell_str);
		goto end2;
		}
    CheckPins(0, 1, 2);
    CheckPins(1, 0, 2);
    CheckPins(0, 2, 1);
    CheckPins(2, 0, 1);
    CheckPins(1, 2, 0);
    CheckPins(2, 1, 0);
    if (((PartFound == 0) || (PartFound == 6) || (PartFound == 1))) 
		{
		EntladePins();
		cap.cval_max = 0;
		cap.cpre_max = -12;
		ReadCapacity(2, 0);
		ReadCapacity(2, 1);
		ReadCapacity(1, 0);
		}
    lcd_clear();
    if (PartFound == 1) {
	if (NumOfDiodes == 1) {
	    lcd_testpin(diodes[0].Anode);
	    lcd_fix_string(AnKat);
	    lcd_testpin(diodes[0].Cathode);
	    UfOutput(0x70);
	    lcd_fix_string(GateCap_str);
	    ReadCapacity(diodes[0].Cathode, diodes[0].Anode);
	    DisplayValue(cap.cval, cap.cpre, 'F', 3);
	    goto end;
	} else if (NumOfDiodes == 2) {
	    lcd_data('2');
	    lcd_fix_string(Diodes);
	    if (diodes[0].Anode == diodes[1].Anode) {
		lcd_testpin(diodes[0].Cathode);
		lcd_fix_string(KatAn);
		lcd_testpin(diodes[0].Anode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[1].Cathode);
		UfOutput(0x01);
		goto end;
	    } else if (diodes[0].Cathode == diodes[1].Cathode) {
		lcd_testpin(diodes[0].Anode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[0].Cathode);
		lcd_fix_string(KatAn);
		lcd_testpin(diodes[1].Anode);
		UfOutput(0x01);
		goto end;
	    } else if ((diodes[0].Cathode == diodes[1].Anode) && (diodes[1].Cathode == diodes[0].Anode)) {
		lcd_testpin(diodes[0].Anode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[0].Cathode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[1].Cathode);
		UfOutput(0x01);
		goto end;
	    }
	} else if (NumOfDiodes == 3) {
	    trans.b = 3;
	    trans.c = 3;
	    if (diodes[0].Cathode == diodes[1].Anode) {
		trans.b = 0;
		trans.c = 1;
	    }
	    if (diodes[0].Anode == diodes[1].Cathode) {
		trans.b = 1;
		trans.c = 0;
	    }
	    if (diodes[0].Cathode == diodes[2].Anode) {
		trans.b = 0;
		trans.c = 2;
	    }
	    if (diodes[0].Anode == diodes[2].Cathode) {
		trans.b = 2;
		trans.c = 0;
	    }
	    if (diodes[1].Cathode == diodes[2].Anode) {
		trans.b = 1;
		trans.c = 2;
	    }
	    if (diodes[1].Anode == diodes[2].Cathode) {
		trans.b = 2;
		trans.c = 1;
	    }
	    if ((trans.b < 3) && (trans.c < 3)) {
		lcd_data('3');
		lcd_fix_string(Diodes);
		lcd_testpin(diodes[trans.b].Anode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[trans.b].Cathode);
		lcd_fix_string(AnKat);
		lcd_testpin(diodes[trans.c].Cathode);
		UfOutput((trans.b << 4) | trans.c);
		goto end;
	    }
	}
    } else if (PartFound == 2) {
	if (PartReady != 0) {
	    if ((trans.hfe[0] > trans.hfe[1])) {
		tmp = trans.c;
		trans.c = trans.e;
		trans.e = tmp;
	    } else {
		trans.hfe[0] = trans.hfe[1];
		trans.uBE[0] = trans.uBE[1];
	    }
	}
	if (PartMode == 1) 
		{
			lcd_fix_string(NPN_str);
		} 
	else 
		{
			lcd_fix_string(PNP_str);
		}
	if (NumOfDiodes > 2) 
	{
	    if (((PartMode == 1) && (trans.c > trans.e)) || ((PartMode != 1) && (trans.c < trans.e))) {
		lcd_fix_string(AnKat);
	    } else {
		lcd_fix_string(KatAn);
	    }
	}
	
	PinLayout('E', 'B', 'C');
	lcd_set_cursor(1, 0);
	lcd_fix_string(hfe_str);
	DisplayValue(trans.hfe[0], 0, 0, 3);
	lcd_space();
	lcd_fix_string(Uf_str);
	DisplayValue(trans.uBE[0], -3, 'V', 3);
	goto end;
    } else if (PartFound == 3) {
	if (PartMode & 1) {
	    lcd_data('P');
	} else {
	    lcd_data('N');
	}
	lcd_data('-');
	tmp = PartMode / 2;
	if (tmp == (4 / 2)) {
	    lcd_data('D');
	}
	if (tmp == (2 / 2)) {
	    lcd_data('E');
	}
	if (tmp == (6 / 2)) {
	    lcd_fix_string(jfet_str);
	} else {
	    lcd_fix_string(mosfet_str);
	}
	PinLayout('S', 'G', 'D');
	if ((NumOfDiodes > 0) && (PartMode < 4)) {
	    if (((PartMode & 1) && (trans.c < trans.e)) || ((!(PartMode & 1)) && (trans.c > trans.e))) {
		lcd_data(1);
	    } else {
		lcd_data(2);
	    }
	}
	lcd_set_cursor(1, 0);
	if (PartMode < 4) {
	    lcd_fix_string(GateCap_str);
	    ReadCapacity(trans.b, trans.e);
	    DisplayValue(cap.cval, cap.cpre, 'F', 3);
	    lcd_fix_string(vt_str);
	} else {
	    lcd_data('I');
	    lcd_data('=');
	    DisplayValue(trans.uBE[1], -5, 'A', 2);
	    lcd_fix_string(Vgs_str);
	}
	DisplayValue(gthvoltage, -3, 'V', 2);
	goto end;
    } else if (PartFound == 5) {
	lcd_fix_string(Thyristor);
	goto gakOutput;
    } else if (PartFound == 4) {
	lcd_fix_string(Triac);
	goto gakOutput;
    } else if (PartFound == 6) {
	if (ResistorsFound == 1) {
	    lcd_testpin(resis[0].rb);
	    lcd_fix_string(Resistor_str);
	    lcd_testpin(resis[0].ra);
	} else {
	    ii = 0;
	    if (resis[1].rx > resis[0].rx)
		ii = 1;
	    if (ResistorsFound == 2) {
		ii = 2;
	    } else {
		if (resis[2].rx > resis[ii].rx)
		    ii = 2;
	    }
	    char x = '1';
	    char y = '3';
	    char z = '2';
	    if (ii == 1) {
		y = '2';
		z = '3';
	    }
	    if (ii == 2) {
		x = '2';
		y = '1';
		z = '3';
	    }
	    lcd_data(x);
	    lcd_fix_string(Resistor_str);
	    lcd_data(y);
	    lcd_fix_string(Resistor_str);
	    lcd_data(z);
	}
	lcd_set_cursor(1, 0);
	if (ResistorsFound == 1) {
	    RvalOut(0);
	} else {
	    if (ii == 0) {
		RvalOut(1);
		RvalOut(2);
	    }
	    if (ii == 1) {
		RvalOut(0);
		RvalOut(2);
	    }
	    if (ii == 2) {
		RvalOut(0);
		RvalOut(1);
	    }
	}
	goto end;
    } else if (PartFound == 7) {
	lcd_testpin(cap.ca);
	lcd_fix_string(CapZeich);
	lcd_testpin(cap.cb);
	lcd_set_cursor(1, 0);
	DisplayValue(cap.cval_max, cap.cpre_max, 'F', 4);
	goto end;
    }
    if (NumOfDiodes == 0) {
	lcd_fix_string(TestFailed1);
	lcd_set_cursor(1, 0);
	lcd_fix_string(TestFailed2);
	lcd_fix_string(Component);
    } else {
	lcd_fix_string(Component);
	lcd_fix_string(Unknown);
	lcd_set_cursor(1, 0);
	lcd_fix_string(OrBroken);
	lcd_data(NumOfDiodes + '0');
	lcd_fix_string(AnKat);
    }
    empty_count++;
    mess_count = 0;
    goto end2;
  gakOutput:
    lcd_set_cursor(1, 0);
    PinLayout('C', 'G', 'A');
  end:
    uart_putc(10);
    //Serial.println("END");
    empty_count = 0;
    mess_count++;
  end2:
    DDRC = 0xF8;
    Serial.println("END");
    goto start;
    while (!(PIND & (1 << 17)));
    sleep_5ms(40);
    for (gthvoltage = 0; gthvoltage < 28000; gthvoltage += 10) {
	if (!(PIND & (1 << 17))) {
	    goto start;
	}
	wdt_reset();
	sleep_5ms(2);
    }
    goto start;
    return;
}






void UfOutput(uint8_t bcdnum)
{
    lcd_set_cursor(1, 0);
    lcd_fix_string(Uf_str);
    mVOutput(bcdnum >> 4);
    mVOutput(bcdnum & 0x0f);
}

void mVOutput(uint8_t nn)
{
    if (nn < 3) 
    {
	DisplayValue(diodes[nn].Voltage, -3, 'V', 3);
	lcd_space();
    }
}

void RvalOut(uint8_t ii)
{
    DisplayValue(resis[ii].rx, -1, 244, 4);
    lcd_space();
}

void ChargePin10ms(uint8_t PinToCharge, uint8_t ChargeDirection)
{
    if (ChargeDirection & 1) 
		{
		PORTB |= PinToCharge;
		} 
    else 
		{
		PORTB &= ~PinToCharge;
		}
    DDRB |= PinToCharge;
    sleep_5ms(2);
    DDRB &= ~PinToCharge;
    PORTB &= ~PinToCharge;
}

void EntladePins()
{
    uint8_t adc_gnd;
    unsigned int adcmv[3];
    unsigned int clr_cnt;
    uint8_t lop_cnt;
    for (lop_cnt = 0; lop_cnt < 10; lop_cnt++) {
	adc_gnd = 0xF8;
	DDRC = adc_gnd;
	PORTC = 0xF8;
	PORTB = 0;
	DDRB = (2 << (2 * 2)) | (2 << (1 * 2)) | (2 << (0 * 2));
	adcmv[0] = W5msReadADC(0);
	adcmv[1] = ReadADC(1);
	adcmv[2] = ReadADC(2);
	if ((PartFound == 8) || (adcmv[0] < 4)
	    & (adcmv[1] < 4)
	    & (adcmv[2] < 4)) {
	    DDRC = 0xF8;
	    DDRB = 0;
	    return;
	}
	if (adcmv[0] < 1000) {
	    adc_gnd |= (1 << 0);
	}
	if (adcmv[1] < 1000) {
	    adc_gnd |= (1 << 1);
	}
	if (adcmv[2] < 1000) {
	    adc_gnd |= (1 << 2);
	}
	DDRC = adc_gnd;
	DDRB = (1 << (2 * 2)) | (1 << (1 * 2)) | (1 << (0 * 2));
	for (clr_cnt = 0; clr_cnt < (10000 / 20); clr_cnt++) {
	    wdt_reset();
	    adcmv[0] = W20msReadADC(0);
	    adcmv[1] = ReadADC(1);
	    adcmv[2] = ReadADC(2);
	    if (adcmv[0] < 1300) {
		DDRC |= (1 << 0);
	    }
	    if (adcmv[1] < 1300) {
		DDRC |= (1 << 1);
	    }
	    if (adcmv[2] < 1300) {
		DDRC |= (1 << 2);
	    }
	    if ((adcmv[0] < (4 + 2)) && (adcmv[1] < (4 + 2)) && (adcmv[2] < (4 + 2))) {
		break;
	    }
	}
	if (clr_cnt == (10000 / 20)) {
	    PartFound = 8;
	}
	for (adcmv[0] = 0; adcmv[0] < clr_cnt; adcmv[0]++) {
	    delay(1);
	}
    }
}

void RefVoltage(void)
{
    unsigned int referenz;
    unsigned int y1, y2;
    uint8_t tabind;
    uint8_t tabres;
    referenz = ref_mv + 12;
    
    if (referenz >= 1000) 
		{
		referenz -= 1000;
		} 
    else 
		{
		referenz = 0;
		}
    tabind = referenz / 50;
    tabres = referenz % 50;
    tabres = 50 - tabres;
    
    if (tabind > 7) {	tabind = 7;    }
    
    y1 = pgm_read_word(&RHtab[tabind]);
    y2 = pgm_read_word(&RHtab[tabind + 1]);
    RHmultip = ((y1 - y2) * tabres + (50 / 2)) / 50 + y2;
}

void DisplayValue(unsigned long Value, int8_t Exponent, unsigned char Unit, unsigned char digits)
{
    char OutBuffer[15];
    unsigned int Limit;
    unsigned char Prefix;
    uint8_t Offset;
    uint8_t Index;
    uint8_t Length;
    Limit = 100;
    if (digits == 3)
	Limit = 1000;
    if (digits == 4)
	Limit = 10000;
    while (Value >= Limit) 
		{
		Value += 5;
		Value = Value / 10;
		Exponent++;
		}
    Length = Exponent + 12;
    if ((int8_t) Length < 0)
	Length = 0;
    if (Length > 18)
	Length = 18;
    Index = Length / 3;
    Offset = Length % 3;
    if (Offset > 0) {
	Index++;
	Offset = 3 - Offset;
    }
    Prefix = pgm_read_byte((uint8_t *) (&PrefixTab[Index]));
    utoa((unsigned int) Value, OutBuffer, 10);
    Length = strlen(OutBuffer);
    Exponent = Length - Offset;
    if (Exponent <= 0) {
	lcd_data('.');
	if (Exponent < 0)
	    lcd_data('0');
    }
    if (Offset == 0)
	Exponent = -1;
    Index = 0;
    while (Index < Length) 
		{
		lcd_data(OutBuffer[Index]);
		Index++;
		if (Index == Exponent) {	    lcd_data('.');	}
		}
    
    if (Prefix != 0)	lcd_data(Prefix);
    if (Unit)	lcd_data(Unit);
}

void sleep_5ms(uint16_t pause)
{
    uint8_t t2_offset;
    while (pause > 0) {
	if (pause > 1) {
	    t2_offset = (10000 - (16384 / (16000000UL / 1000000UL))) / (1024 / (16000000UL / 1000000UL));;
	    pause -= 2;
	} else {
	    t2_offset = (5000 - (16384 / (16000000UL / 1000000UL))) / (1024 / (16000000UL / 1000000UL));;
	    pause = 0;
	}
	OCR2A = TCNT2 + t2_offset;
	TIMSK2 = (0 << OCIE2B) | (1 << OCIE2A) | (0 << TOIE2);
	set_sleep_mode(SLEEP_MODE_PWR_SAVE);
	sleep_mode();
	wdt_reset();
    }
    TIMSK2 = (0 << OCIE2B) | (0 << OCIE2A) | (0 << TOIE2);
}

void PinLayout(char pin1, char pin2, char pin3)
{
    lcd_fix_string(N123_str);
    for (ii = 0; ii < 3; ii++) 
		{
			if (ii == trans.e)	    lcd_data(pin1);
			if (ii == trans.b)	    lcd_data(pin2);
			if (ii == trans.c)	    lcd_data(pin3);
		}
}

void AutoCheck(void)
{
}

void CheckPins(uint8_t HighPin, uint8_t LowPin, uint8_t TristatePin)
{
    struct {
	unsigned int lp_otr;
	unsigned int hp1;
	unsigned int hp2;
	unsigned int hp3;
	unsigned int lp1;
	unsigned int lp2;
	unsigned int tp1;
	unsigned int tp2;
    } adc;
    uint8_t LoPinRL;
    uint8_t LoPinRH;
    uint8_t TriPinRL;
    uint8_t TriPinRH;
    uint8_t HiPinRL;
    uint8_t HiPinRH;
    uint8_t HiADCp;
    uint8_t LoADCp;
    uint8_t HiADCm;
    uint8_t LoADCm;
    uint8_t PinMSK;
    uint8_t ii;
    unsigned long c_hfe;
    struct resis_t *thisR;
    unsigned long lrx1;
    unsigned long lirx1;
    unsigned long lirx2;
    wdt_reset();
    LoPinRL = pgm_read_byte(&PinRLtab[LowPin]);
    LoPinRH = LoPinRL + LoPinRL;
    TriPinRL = pgm_read_byte(&PinRLtab[TristatePin]);
    TriPinRH = TriPinRL + TriPinRL;
    HiPinRL = pgm_read_byte(&PinRLtab[HighPin]);
    HiPinRH = HiPinRL + HiPinRL;
    HiADCp = pgm_read_byte(&PinADCtab[HighPin]);
    LoADCp = pgm_read_byte(&PinADCtab[LowPin]);
    HiADCm = HiADCp | 0xF8;
    HiADCp |= 0xF8;
    LoADCm = LoADCp | 0xF8;
    LoADCp |= 0xF8;
    PORTB = 0;
    DDRB = LoPinRL;
    DDRC = HiADCm;
    PORTC = HiADCp;
    ChargePin10ms(TriPinRL, 0);
    adc.lp_otr = W5msReadADC(LowPin);
    if (adc.lp_otr >= 977) {
	ChargePin10ms(TriPinRL, 1);
	adc.lp_otr = ReadADC(LowPin);
    }
    if (adc.lp_otr > 455) {
	DDRB = LoPinRL | TriPinRH;
	adc.lp1 = W20msReadADC(LowPin);
	adc.tp1 = ReadADC(TristatePin);
	PORTB = TriPinRH;
	adc.lp2 = W20msReadADC(LowPin);
	if (adc.lp2 > (adc.lp1 + 488)) {
	    if (PartFound != 3) {
		PORTC = 0xF8;
		DDRC = LoADCm;
		DDRB = TriPinRH | HiPinRL;
		PORTB = TriPinRH | HiPinRL;
		adc.lp2 = W20msReadADC(TristatePin);
		if (adc.lp2 > 3911) {
		    PartFound = 3;
		    PartMode = 4;
		} else {
		    PartFound = 3;
		    PartMode = 6;
		}
		trans.uBE[0] = adc.lp1;
		gthvoltage = adc.lp1 - adc.tp1;
		trans.uBE[1] = (unsigned int) (((unsigned long) adc.lp1 * 1000) / (6800 + 200));
		trans.b = TristatePin;
		trans.c = HighPin;
		trans.e = LowPin;
	    }
	}
	PORTC = 0xF8;
	DDRC = LoADCm;
	DDRB = TriPinRH | HiPinRL;
	PORTB = TriPinRH | HiPinRL;
	adc.hp1 = W20msReadADC(HighPin);
	adc.tp1 = ReadADC(TristatePin);
	PORTB = HiPinRL;
	adc.hp2 = W20msReadADC(HighPin);
	if (adc.hp1 > (adc.hp2 + 488)) {
	    if (PartFound != 3) {
		PORTC = HiADCp;
		DDRC = HiADCm;
		adc.tp2 = W20msReadADC(TristatePin);
		if (adc.tp2 < 977) {
		    PartFound = 3;
		    PartMode = 5;
		} else {
		    PartFound = 3;
		    PartMode = 7;
		}
		gthvoltage = adc.tp1 - adc.hp1;
		trans.uBE[1] = (unsigned int) (((unsigned long) (ADCconfig.U_AVCC - adc.hp1) * 1000) / (6800 + 220));
		trans.b = TristatePin;
		trans.c = LowPin;
		trans.e = HighPin;
	    }
	}
    }
    PORTC = 0xF8;
    DDRC = LoADCm;
    PORTB = HiPinRL;
    DDRB = TriPinRL | HiPinRL;
    adc.hp1 = ADCconfig.U_AVCC - W5msReadADC(HighPin);
    adc.tp1 = ReadADC(TristatePin);
    if (adc.tp1 < 10) {
	DDRB = TriPinRH | HiPinRL;
	adc.hp1 = ADCconfig.U_AVCC - W5msReadADC(HighPin);
	adc.tp1 = ReadADC(TristatePin);
	c_hfe = ((unsigned long) adc.hp1 * (unsigned long) (((unsigned long) 47000 * 100) / (unsigned int) (6800 + 220))) / (unsigned int) adc.tp1;
    } else {
	c_hfe = (unsigned long) ((adc.hp1 - adc.tp1) / adc.tp1);
    }
    DDRB = LoPinRL;
    PORTB = 0;
    DDRC = HiADCm;
    PORTC = HiADCp;
    sleep_5ms(1);
    if (adc.lp_otr < 977) {
	DDRB = LoPinRL | TriPinRL;
	adc.lp1 = W5msReadADC(LowPin);
	if (adc.lp1 > 3422) {
	    DDRB = LoPinRL | TriPinRH;
	    adc.lp1 = W5msReadADC(LowPin);
	    adc.tp2 = ReadADC(TristatePin);
	    if ((PartFound == 2) || (PartFound == 3)) {
		PartReady = 1;
	    }
	    trans.hfe[PartReady] = c_hfe;
	    trans.uBE[PartReady] = ADCconfig.U_AVCC - adc.hp1 - adc.tp1;
	    if (PartFound != 5) {
		if (adc.tp2 > 977) {
		    PartFound = 2;
		    PartMode = 2;
		} else {
		    if ((adc.lp_otr < 97) && (adc.lp1 > 2000)) {
			PartFound = 3;
			PartMode = 3;
			PinMSK = LoADCm & 7;
			ADMUX = TristatePin | (1 << REFS0);
			gthvoltage = 1;
			for (ii = 0; ii < 11; ii++) {
			    wdt_reset();
			    ChargePin10ms(TriPinRL, 1);
			    DDRB = LoPinRL | TriPinRH;
			    while (!(PINC & PinMSK));
			    DDRB = LoPinRL;
			    ADCSRA |= (1 << ADSC);
			    while (ADCSRA & (1 << ADSC));
			    gthvoltage += (1023 - ADCW);
			}
			gthvoltage *= 4;
			gthvoltage /= 9;
		    }
		}
		trans.b = TristatePin;
		trans.c = LowPin;
		trans.e = HighPin;
	    }
	}
	DDRB = LoPinRL | TriPinRL;
	PORTB = TriPinRL;
	adc.lp1 = W5msReadADC(LowPin);
	adc.tp1 = ADCconfig.U_AVCC - ReadADC(TristatePin);
	if (adc.tp1 < 10) {
	    DDRB = LoPinRL | TriPinRH;
	    PORTB = TriPinRH;
	    adc.lp1 = W5msReadADC(LowPin);
	    adc.tp1 = ADCconfig.U_AVCC - ReadADC(TristatePin);
	    c_hfe = ((unsigned long) adc.lp1 * (unsigned long) (((unsigned long) 47000 * 100) /
								(unsigned int) (6800 + 200))) / (unsigned int) adc.tp1;
	} else {
	    c_hfe = (adc.lp1 - adc.tp1) / adc.tp1;
	}
	DDRC = LoADCm;
	PORTC = 0xF8;
	DDRB = TriPinRL | HiPinRL;
	PORTB = TriPinRL | HiPinRL;
	adc.hp1 = W5msReadADC(HighPin);
	if (adc.hp1 < 1600) {
	    if (PartReady == 1) {
		goto widmes;
	    }
	    ChargePin10ms(TriPinRL, 0);
	    adc.hp3 = W5msReadADC(HighPin);
	    PORTB = 0;
	    sleep_5ms(1);
	    PORTB = HiPinRL;
	    adc.hp2 = W5msReadADC(HighPin);
	    if ((adc.hp3 < 1600) && (adc.hp2 > 4400)) {
		PartFound = 5;
		DDRB = 0;
		PORTB = 0;
		PORTC = LoADCp;
		sleep_5ms(1);
		DDRB = HiPinRL;
		if (W5msReadADC(HighPin) > 244) {
		    goto savenresult;
		}
		DDRB = HiPinRL | TriPinRL;
		if (W5msReadADC(TristatePin) < 977) {
		    goto savenresult;
		}
		if (ReadADC(HighPin) < 733) {
		    goto savenresult;
		}
		DDRB = HiPinRL;
		if (W5msReadADC(HighPin) < 733) {
		    goto savenresult;
		}
		PORTB = HiPinRL;
		sleep_5ms(1);
		PORTB = 0;
		if (W5msReadADC(HighPin) > 244) {
		    goto savenresult;
		}
		PartFound = 4;
		PartReady = 1;
		goto savenresult;
	    }
	    DDRB = HiPinRL | TriPinRH;
	    PORTB = HiPinRL | TriPinRH;
	    sleep_5ms(10);
	    adc.hp2 = ADCconfig.U_AVCC - ReadADC(HighPin);
	    adc.tp2 = ADCconfig.U_AVCC - ReadADC(TristatePin);
	    if ((PartFound == 2) || (PartFound == 3)) {
		PartReady = 1;
	    }
	    trans.hfe[PartReady] = c_hfe;
	    trans.uBE[PartReady] = ADCconfig.U_AVCC - adc.lp1 - adc.tp1;
	    if (adc.tp2 > 2557) {
		PartFound = 2;
		PartMode = 1;
	    } else {
		if ((adc.lp_otr < 97) && (adc.hp2 > 3400)) {
		    PartFound = 3;
		    PartMode = 2;
		    PinMSK = HiADCm & 7;
		    ADMUX = TristatePin | (1 << REFS0);
		    gthvoltage = 1;
		    for (ii = 0; ii < 11; ii++) {
			wdt_reset();
			ChargePin10ms(TriPinRL, 0);
			DDRB = HiPinRL | TriPinRH;
			PORTB = HiPinRL | TriPinRH;
			while ((PINC & PinMSK));
			DDRB = HiPinRL;
			ADCSRA |= (1 << ADSC);
			while (ADCSRA & (1 << ADSC));
			gthvoltage += ADCW;
		    }
		    gthvoltage *= 4;
		    gthvoltage /= 9;
		}
	    }
	  savenresult:
	    trans.b = TristatePin;
	    trans.c = HighPin;
	    trans.e = LowPin;
	}
	DDRC = 0xF8;
	PORTC = 0xF8;
	goto widmes;
    }
    PORTC = 0xF8;
    for (ii = 0; ii < 200; ii++) {
	DDRC = LoADCm | HiADCm;
	sleep_5ms(1);
	DDRC = LoADCm;
	adc.hp1 = ReadADC(HighPin);
	if (adc.hp1 < (150 / 8))
	    break;
    }
    DDRB = HiPinRH;
    PORTB = HiPinRH;
    ChargePin10ms(TriPinRL, 1);
    adc.hp2 = W5msReadADC(HighPin);
    ChargePin10ms(TriPinRL, 0);
    adc.hp3 = W5msReadADC(HighPin);
    DDRB = HiPinRL;
    PORTB = HiPinRL;
    adc.hp1 = W5msReadADC(HighPin) - ReadADC(LowPin);
    ChargePin10ms(TriPinRL, 1);
    adc.lp_otr = W5msReadADC(HighPin) - ReadADC(LowPin);
    DDRB = HiPinRH;
    PORTB = HiPinRH;
    if (adc.lp_otr > adc.hp1) {
	adc.hp1 = adc.lp_otr;
	adc.hp3 = adc.hp2;
    } else {
	ChargePin10ms(TriPinRL, 0);
    }
    adc.hp2 = W5msReadADC(HighPin);
    if ((adc.hp1 > 150) && (adc.hp1 < 4640) && (adc.hp2 < adc.hp1) && (adc.hp1 > (adc.hp3 + (adc.hp3 / 8))) && (adc.hp3 * 16 > adc.hp1)) {
	if ((PartFound == 0) || (PartFound == 6)) {
	    PartFound = 1;
	}
	diodes[NumOfDiodes].Anode = HighPin;
	diodes[NumOfDiodes].Cathode = LowPin;
	diodes[NumOfDiodes].Voltage = adc.hp1;
	NumOfDiodes++;
    }
  widmes:
    if (NumOfDiodes > 0)
	goto clean_ports;
    wdt_reset();
    ADCconfig.U_AVCC *= 4;
    ADCconfig.U_Bandgap *= 4;
    ADCconfig.Samples = 190;
    PORTC = 0xF8;
    DDRC = LoADCm;
    DDRB = HiPinRL;
    PORTB = HiPinRL;
    adc.tp1 = W5msReadADC(LowPin);
    adc.hp1 = ReadADC(HighPin);
    if (adc.tp1 > adc.hp1) {
	adc.tp1 = adc.hp1;
    }
    PORTB = 0;
    DDRB = HiPinRH;
    adc.hp2 = W5msReadADC(HighPin);
    if (adc.hp2 > (20 * 4)) {
	goto testend;
    }
    PORTB = HiPinRH;
    adc.hp2 = W5msReadADC(HighPin);
    DDRC = HiADCm;
    PORTC = HiADCp;
    PORTB = 0;
    DDRB = LoPinRL;
    adc.tp2 = W5msReadADC(HighPin);
    adc.lp1 = ReadADC(LowPin);
    if (adc.tp2 < adc.lp1) {
	adc.tp2 = adc.lp1;
    }
    DDRB = LoPinRH;
    adc.lp2 = W5msReadADC(LowPin);
    if ((adc.hp1 < (4400 * 4)) && (adc.hp2 > (97 * 4))) {
	goto testend;
    }
    if (adc.hp2 < (4972 * 4)) {
	if (adc.lp1 < (169 * 4)) {
	    ii = 'H';
	    if (adc.lp2 < (38 * 4)) {
		goto testend;
	    }
	    lirx1 = (unsigned long) ((unsigned int) 47000) * (unsigned long) adc.hp2 / (ADCconfig.U_AVCC - adc.hp2);
	    lirx2 = (unsigned long) ((unsigned int) 47000) * (unsigned long) (ADCconfig.U_AVCC - adc.lp2) / adc.lp2;
	    if (adc.hp2 < (990 * 4)) {
		lrx1 = (lirx1 * 4 + lirx2) / (4 + 1);
	    } else if (adc.lp2 < (990 * 4)) {
		lrx1 = (lirx2 * 4 + lirx1) / (4 + 1);
	    } else {
		lrx1 = (lirx1 + lirx2) / 2;
	    }
	    lrx1 *= 100;
	    lrx1 += 3500;
	} else {
	    ii = 'L';
	    if (adc.tp1 > adc.hp1) {
		adc.hp1 = adc.tp1;
	    }
	    lirx1 = (unsigned long) (6800 + 220) * (unsigned long) (adc.hp1 - adc.tp1) / (ADCconfig.U_AVCC - adc.hp1);
	    if (adc.tp2 < adc.lp1) {
		adc.lp1 = adc.tp2;
	    }
	    lirx2 = (unsigned long) (6800 + 200) * (unsigned long) (adc.tp2 - adc.lp1) / adc.lp1;
	    if (adc.hp1 < (990 * 4)) {
		lrx1 = (lirx1 * 4 + lirx2) / (4 + 1);
	    } else if (adc.lp1 < (990 * 4)) {
		lrx1 = (lirx2 * 4 + lirx1) / (4 + 1);
	    } else {
		lrx1 = (lirx1 + lirx2) / 2;
	    }
	}
	if ((PartFound == 1) || (PartFound == 0) || (PartFound == 6)) {
	    for (ii = 0; ii < ResistorsFound; ii++) {
		thisR = &resis[ii];
		if (thisR->rt != TristatePin)
		    continue;
		lirx1 = (labs((long) lrx1 - (long) thisR->rx) * 10) / (lrx1 + thisR->rx + 100);
		if (lirx1 > 0) {
		    goto testend;
		}
		PartFound = 6;
		goto testend;
	    }
	    thisR = &resis[ResistorsFound];
	    thisR->rx = lrx1;
	    thisR->ra = LowPin;
	    thisR->rb = HighPin;
	    thisR->rt = TristatePin;
	    ResistorsFound++;
	}
    }
  testend:
    ADCconfig.U_AVCC /= 4;
    ADCconfig.U_Bandgap /= 4;
    ADCconfig.Samples = 25;
  clean_ports:
    DDRC = 0xF8;
    PORTC = 0xF8;
    DDRB = 0;
    PORTB = 0;
}


void GetIr(uint8_t hipin, uint8_t lopin)
{
    unsigned int u_res;
    unsigned int ir_nano;
    uint8_t LoPinR_L;
    uint8_t HiADC;
    HiADC = pgm_read_byte(&PinADCtab[hipin]);
    PORTC = HiADC | 0xF8;
    DDRC = HiADC | 0xF8;
    LoPinR_L = pgm_read_byte(&PinRLtab[lopin]);
    PORTB = 0;
    DDRB = LoPinR_L + LoPinR_L;
    u_res = W5msReadADC(lopin);
    if (u_res == 0)
	return;
    lcd_fix_string(Ir_str);
    ir_nano = (unsigned long) (u_res * 100000UL) / 47000;
    DisplayValue(ir_nano, -9, 'A', 2);
    DDRC = 0xF8;
    PORTC = 0xF8;
    DDRB = 0;
    return;


}



unsigned int ReadADC(uint8_t Probe)
{
	unsigned int U;
	uint8_t Samples;
	unsigned long Value;
	Probe |= (1 << REFS0);
  sample:
	ADMUX = Probe;
	if ((Probe & (1 << REFS1)) != 0) {		delayMicroseconds(100);	}
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
	set_sleep_mode(SLEEP_MODE_ADC);
	sleep_mode();;
	Value = 0UL;
	Samples = 0;
	while (Samples < ADCconfig.Samples) 
	{
		ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADIE) | (1 << ADPS2) | (1 << ADPS1)	| (1 << ADPS0);
		set_sleep_mode(SLEEP_MODE_ADC);
		sleep_mode();;
		Value += ADCW;
		if ((Samples == 4) && (ADCconfig.U_Bandgap > 255)	&& ((uint16_t) Value < 1024) && !(Probe & (1 << REFS1))) 
		{
			Probe |= (1 << REFS1);
			goto sample;
		}
		Samples++;
	}
	if (Probe & (1 << REFS1))
		U = ADCconfig.U_Bandgap;
	else
		U = ADCconfig.U_AVCC;
	Value *= U;
	Value /= 1023;
	Value /= ADCconfig.Samples;
	U = (unsigned int) Value;
	return U;
}



unsigned int W5msReadADC(uint8_t Probe)
{
    sleep_5ms(1);
    return (ReadADC(Probe));
}

unsigned int W10msReadADC(uint8_t Probe)
{
    sleep_5ms(2);
    return (ReadADC(Probe));
}

unsigned int W20msReadADC(uint8_t Probe)
{
    sleep_5ms(4);
    return (ReadADC(Probe));
}

void ReadCapacity(uint8_t HighPin, uint8_t LowPin)
{
    unsigned int tmpint;
    unsigned int adcv[4];
    uint8_t HiPinR_L, HiPinR_H;
    uint8_t LoADC;
    uint8_t ii;
    LoADC = pgm_read_byte(&PinADCtab[LowPin]) | 0xF8;
    HiPinR_L = pgm_read_byte(&PinRLtab[HighPin]);
    HiPinR_H = HiPinR_L + HiPinR_L;
    if (PartFound == 6) {	return;    }
    
    for (ii = 0; ii < NumOfDiodes; ii++) 
		{
			if ((diodes[ii].Cathode == LowPin) && (diodes[ii].Anode == HighPin) && (diodes[ii].Voltage < 1500)) { return;	}
		}
    cap.cval = 0;
    cap.cpre = -12;
    EntladePins();
    PORTC = 0xF8;
    PORTB = 0;
    DDRC = LoADC;
    DDRB = HiPinR_L;
    adcv[0] = ReadADC(HighPin);
    adcv[2] = adcv[0];
    for (ovcnt16 = 0; ovcnt16 < 500; ovcnt16++) 
    {
		PORTB = HiPinR_L;
		DDRB = HiPinR_L;
		delay(10);
		DDRB = 0;
		PORTB = 0;
		delayMicroseconds(500);
		wdt_reset();
		adcv[2] = ReadADC(HighPin);
		if (adcv[2] > adcv[0]) 
			{
				adcv[2] -= adcv[0];
			} 
			else 
			{
				adcv[2] = 0;
			}
		if ((ovcnt16 == 126) && (adcv[2] < 75)) {	    break;	}
		if (adcv[2] > 300) {	    break;	}
    }
    
    if (adcv[2] < 301) {	goto keinC;    }
    if ((ovcnt16 == 0) && (adcv[2] > 1300)) {	goto messe_mit_rh;    }
    adcv[3] = adcv[2];
    for (tmpint = 0; tmpint <= ovcnt16; tmpint++) 
		{
		delay(5);
		adcv[3] = ReadADC(HighPin);
		wdt_reset();
		}
    if (adcv[3] > adcv[0]) {	adcv[3] -= adcv[0];    } 
		else {	adcv[3] = 0;    }
		
    if (adcv[2] > adcv[3]) 
		{
		adcv[3] = adcv[2] - adcv[3];
		} 
	 else 
		{
		adcv[3] = 0;
		}
    if (adcv[3] > 100) 
		{
			if (ovcnt16 == 0) {	    goto messe_mit_rh;	}
			goto keinC;
		}
    cap.cval_uncorrected.dw = ovcnt16 + 1;
    cap.cval_uncorrected.dw *= getRLmultip(adcv[2] + adcv[3]);
    cap.cval = cap.cval_uncorrected.dw;
    cap.cpre = -9;
    Scale_C_with_vcc();
    cap.cval *= (1000 - 0);
    cap.cval /= 100;
    goto checkDiodes;
  messe_mit_rh:
    EntladePins();
    PORTB = 0;
    DDRC = (1 << 0) | (1 << 1) | (1 << 2) | (1 << 3);
    PORTC = 0xF8;
    DDRB = HiPinR_H;
    ADCSRB = (1 << ACME);
    ACSR = (1 << ACBG) | (1 << ACI) | (1 << ACIC);
    ADMUX = (1 << REFS0) | HighPin;
    ADCSRA = (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
    delayMicroseconds(200);
    ovcnt16 = 0;
    TCCR1A = 0;
    TCNT1 = 0;
    TIFR1 = (1 << ICF1) | (1 << OCF1B) | (1 << OCF1A) | (1 << TOV1);
    TIMSK1 = (1 << TOIE1) | (1 << ICIE1);
    unfinished = 1;
    PORTB = HiPinR_H;
    if (PartFound == 3) 
		{
			TCCR1B = (1 << CS10);
			DDRC = (((1 << 0) | (1 << 1) | (1 << 2) | 0xF8) & ~(1 << HighPin));
		} 
		else 
		{
			TCCR1B = (1 << CS10);
			DDRC = LoADC;
		}
		
    while (unfinished) 
		{
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();
		wdt_reset();
		if (ovcnt16 == (16000000UL / 5000)) {    break;	}
		}
    TCCR1B = (0 << ICNC1) | (0 << ICES1) | (0 << CS10);
    tmpint = ICR1;
    TIMSK1 = (0 << TOIE1) | (0 << ICIE1);
    if (TCNT1 < tmpint) {
	ovcnt16--;
    }
    ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
    DDRB = 0;
    PORTB = 0;
    adcv[2] = ReadADC(HighPin);
    load_diff = adcv[2] + 12 - ref_mv;
    if (ovcnt16 >= (16000000UL / 10000)) {
	goto keinC;
    }
    cap.cval_uncorrected.w[1] = ovcnt16;
    cap.cval_uncorrected.w[0] = tmpint;
    cap.cpre = -12;
    if (ovcnt16 > 65) {
	cap.cval_uncorrected.dw /= 100;
	cap.cpre += 2;
    }
    cap.cval_uncorrected.dw *= RHmultip;
    cap.cval_uncorrected.dw /= (16000000UL / 10000);
    cap.cval = cap.cval_uncorrected.dw;
    Scale_C_with_vcc();
    if (cap.cpre == -12) {
	if (cap.cval < 4000) {
	    cap.cval += (4000 / (cap.cval + 180));
	}
	if (HighPin == 1)
	    cap.cval += 2;
	if (cap.cval > 36 + 3 + (4000 / (36 + 3 + 180))) {
	    cap.cval -= 36 + 3 + (4000 / (36 + 3 + 180));
	} else {
	    cap.cval = 0;
	}
    }
    DDRB = HiPinR_L;
    if (cap.cval < 25) {
	goto keinC;
    }
  checkDiodes:
    if ((NumOfDiodes > 0) && (PartFound != 3)) 
    {
    } 
    else 
    {
		PartFound = 7;
		if ((cap.cpre > cap.cpre_max) || ((cap.cpre == cap.cpre_max) && (cap.cval > cap.cval_max)))
			{
			cap.cval_max = cap.cval;
			cap.cpre_max = cap.cpre;
			cap.ca = LowPin;
			cap.cb = HighPin;
			}
    }
  keinC:
    DDRC = 0xF8;
    PORTC = 0xF8;
    DDRB = 0;
    PORTB = 0;
    return;
}

unsigned int getRLmultip(unsigned int cvolt)
{
    unsigned int uvolt;
    unsigned int y1, y2;
    uint8_t tabind;
    uint8_t tabres;
    if (cvolt >= 300) 
		{
			uvolt = cvolt - 300;
		}
		 else 
		{
			uvolt = 0;
		}
    tabind = uvolt / 25;
    tabres = uvolt % 25;
    tabres = 25 - tabres;
    if (tabind > (1100 / 25)) {	tabind = (1100 / 25);    }
    y1 = pgm_read_word(&RLtab[tabind]);
    y2 = pgm_read_word(&RLtab[tabind + 1]);
    return (((y1 - y2) * tabres + (25 / 2)) / 25 + y2);
}

void Scale_C_with_vcc(void)
{
    while (cap.cval > 100000) 
      {
  	cap.cval /= 10;
  	cap.cpre++;
      }
    cap.cval *= ADCconfig.U_AVCC;
    cap.cval /= 5000;
}

ISR(TIMER1_OVF_vect, ISR_BLOCK)
{
    ovcnt16++;
}

ISR(TIMER1_CAPT_vect, ISR_BLOCK)
{
    unfinished = 0;
}

void ReadInductance(void)
{
    return;
}

uint16_t GetESR(uint8_t hipin, uint8_t lopin)
{
    return (0);
}

void us500delay(unsigned int us)
{
    if (--us == 0)
	return;
    us <<= 2;
    __asm__ __volatile__("1: sbiw %0,1" "\n\t" "brne 1b":"=w"(us):"0"(us) );
}

void GetVloss()
{
    return;
}

void Calibrate_UR(void)
{
    ADCconfig.U_AVCC = 5000;
    ADCconfig.Samples = 190;
    (void) ReadADC(0x0e);
    ref_mv = W5msReadADC(0x0e);
    RefVoltage();
    ADCconfig.U_Bandgap = (ref_mv + 3);
    ADCconfig.Samples = 25;
}

void lcd_set_cursor(uint8_t row, uint8_t col)
{
    int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
    if (row >= 2) {	row = 1;  }
    Serial.println();
}

void lcd_string(char *data)
{
    while (*data) 
    {
		lcd_data(*data);
		data++;
    }
}

void lcd_pgm_string(const unsigned char *data)
{
    unsigned char cc;
    while (1) 
    {
		cc = pgm_read_byte(data);
		if ((cc == 0) || (cc == 128))    return;
		lcd_data(cc);
		data++;
    }
}

void lcd_pgm_custom_char(uint8_t location, const unsigned char *chardata)
{
    location &= 0x7;
    for (uint8_t i = 0; i < 8; i++) {	chardata++;    }
}


void lcd_testpin(unsigned char temp)
{
    lcd_data(temp + '1');
}


void lcd_space(void)
{
    lcd_data(' ');
}


void lcd_fix_string(const unsigned char *data)
{
    unsigned char cc;
    while (1) 
    {
		cc = pgm_read_byte(data);
		if ((cc == 0) || (cc == 128))    return;
		lcd_data(cc);
		data++;
    }

}


void lcd_data(unsigned char temp1)
{
    switch (temp1) {
    case 1:{
	    uart_putc('>');
	    uart_putc('|');
	    break;
	}
    case 2:{
	    uart_putc('|');
	    uart_putc('<');
	    break;
	}
    case 3:{
	    uart_putc('|');
	    uart_putc('|');
	    break;
	}
    case 7:{
	    uart_putc('[');
	    uart_putc('=');
	    break;
	}
    case 6:{
	    uart_putc(']');
	    break;
	}
    case 228:{
	    uart_putc('u');
	    break;
	}
    case 244:{
	    uart_putc('o');
	    uart_putc('h');
	    uart_putc('m');
	    break;
	}
    default:{
	    uart_putc(temp1);
	}
    }
}


void lcd_clear(void)
{
    //Serial.println();
}


void uart_putc(uint8_t data)
{
    Serial.write(data);
    delay(2);
}
